/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <bio.h>
#include <json.h>
#include "jira.h"

void
warn(char *fmt, ...)
{
	va_list ap;

	fprint(2, "%s: ", argv0);
	va_start(ap, fmt);
	vfprint(2, fmt, ap);
	va_end(ap);
}

char *
strlwr(char *str)
{
	char *s;

	for(s = str; *s; s++)
		*s = tolower(*s);
	return str;
}

char *
strupr(char *str)
{
	char *s;

	for(s = str; *s; s++)
		*s = toupper(*s);
	return str;
}

Jvalue *
exch(Sess *s, char *post, char *fmt, ...)
{
	double then;
	va_list ap;
	int e, fd;
	Jvalue *v;
	Biobuf *bp;
	char *url, *a[4], buf[1024], err[ERRMAX];
	
	then = nsec();
	va_start(ap, fmt);
	url = vsmprint(fmt, ap);
	va_end(ap);

	if(Dumptime || Dumpjson)
		fprint(2, "request: %s/%q\n", s->url, url);

	/* set headers */
	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid);
	fd = open(buf, OWRITE);
	assert(fd != -1);

	if(post){
		fprint(fd, "request POST");
		fprint(fd, "contenttype application/json");
	}

	fprint(fd, "url %s/%s", s->url, url);
	close(fd);

	if(post){
		/* send post body */
		snprint(buf, sizeof buf, "%s/%d/postbody", s->webfs, s->conid);
		fd = open(buf, OWRITE);
		assert(fd != -1);
		fprint(fd, "%s", post);
		close(fd);
	}

	/* read reply */
	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid);
	bp = Bopen(buf, OREAD);

	if(bp == nil){
		rerrstr(err, sizeof(err));
		if(tokenize(err, a, nelem(a)) >= 2)
			if((e = atoi(a[1])) > 0)
				httperrstr(e);
		free(url);
		return nil;
	}

	if((v = json_Bparse(bp, nil)) == nil)
		sysfatal("%s - JSON parse error - %r\n", url);
	Bterm(bp);

	if(Dumptime)
		fprint(2, "%.3f %s\n", (nsec() - then) / 1e9, url);
	if(Dumpjson)
		json_print(2, v);

	free(url);
	return v;
}

char *
fetch(Sess *s, char *url, vlong len)
{
	int fd, got, e, n;
	char *buf, tmp[128], err[ERRMAX], *a[8];
	
	if(len <= 0)
		return estrdup9p("");

	snprint(tmp, sizeof tmp, "%s/%d/ctl", s->webfs, s->conid);
	fd = open(tmp, OWRITE);
	assert(fd != -1);

	fprint(fd, "url %s", url);
	close(fd);

	snprint(tmp, sizeof tmp, "%s/%d/body", s->webfs, s->conid);
	fd = open(tmp, OREAD);
	if(fd < 0){
		rerrstr(err, sizeof(err));
		if(tokenize(err, a, nelem(a)) >= 2)
			if((e = atoi(a[1])) > 0)
				httperrstr(e);
		return nil;
	}

	got = 0;
	buf = emalloc9p(len);
	while((n = read(fd, buf+got, len-got)) > 0){
			got += n;
			if(got > len){
				warn("%s growing file, truncated\n", url);
				break;
			}
	}
	close(fd);
	return buf;
}

int
dialweb(Sess *s)
{
	int n, fd;
	char buf[128];

	snprint(buf, sizeof(buf), "%s/clone", s->webfs);
	if((fd = open(buf, ORDWR)) < 0)
		return -1;
	if((n = read(fd, buf, sizeof buf-1)) <= 0){
		close(fd);
		return -1;
	}
	buf[n] = '\0';
	if ((s->conid = atoi(buf)) <0){
		close(fd);
		return -1;
	}
	s->fd = fd;

	if(s->annon)
		return 0;

	/*
	 * I don't know the relevance of the name
	 * "protected-area" but from inspection,
	 * this is the realm jira wants.
	 */
	snprint(buf, sizeof(buf), "%s/ctl", s->webfs);
	if((fd = open(buf, ORDWR)) < 0)
		return -1;
	fprint(fd, "preauth %s protected-area", s->url);
	close(fd);

	return 0;
}

long
iso2sec(char *s)
{
	Tm tm;
	long t;
	char *td[4], *ymd[5], *hms[5], *tz;

	/* e.g.
	 * 2010-04-06T11:39:46Z
	 * 2011-11-27t00:00:00.000+0000
	 */
	memset(&tm, 0, sizeof(tm));

	if(gettokens(s, td, nelem(td), "tT") != 2)
		return -1L;
	if(gettokens(td[0], ymd, nelem(ymd), "-") != 3)
		return -1L;
	if(gettokens(td[1], hms, nelem(hms), ":") != 2)

	tm.year = strtol(ymd[0], nil, 10) - 1900;
	tm.mon = strtol(ymd[1], nil, 10) -1;
	tm.mday = strtol(ymd[2], nil, 10);
	tm.hour = strtol(hms[0], nil, 10);
	tm.min = strtol(hms[1], nil, 10);
	tm.sec = strtol(hms[2], &tz, 10);
	
	if(*tz == '.')
		strtol(tz+1, &tz, 10);	/* skip fractional secconds */

	if(*tz == 'Z' || *tz == 'z')
		tm.tzoff = 0;
	else
	if(*tz == '+' || *tz == '-')
		tm.tzoff = strtol(tz, nil, 10);

	if((t = tm2sec(&tm)) < 0)
		return -1L;
	return t;
}
