/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <pool.h>
#include <bio.h>
#include <json.h>
#include "jira.h"

typedef struct Aux Aux;
struct Aux {
	Node *node;

	char *buf;	/* write buffer */
	int len;	/* length of buffer in use */
	int siz;	/* buffer size allocated */
};

static struct  {
	Sess *session;
	Node *root;
	long cachetime;
} Params;

int Dumpjson;
int Dumptree;
int Dumptime;
int Dumpcache;
int Dumpvalfail;
char *Debug;

long Userupdate;

static void
dumptree(Node *n, int in)
{

	for (; n; n = n->next){
		print("%*s %D\n", in, "", &n->dir);
		if (n->child)
			dumptree(n->child, in+2);
	}
}

static void
dircp(Dir *d, Dir *s)
{
	*d = *s;
	d->uid = estrdup9p(d->uid);
	d->gid = estrdup9p(d->gid);
	d->muid = estrdup9p(d->muid);
	d->name = estrdup9p(d->name);
}

static uint
hash(char *s)
{
	uint h;
	uchar *p;

	h = 0;
	for(p = (uchar*)s; *p; p++)
		h = h*37 + *p;
	return h;
}

void
scanhook(Node *n, Getfunc func, char *arg)
{
	n->scan.func = func;
	if(arg)
		n->scan.arg = estrdup9p(arg);
}

void
walkhook(Node *n, Getfunc func, char *arg)
{
	n->walk.func = func;
	if(arg)
		n->walk.arg = estrdup9p(arg);
}

void
writehook(Node *n, Putfunc func, char *arg)
{
	n->write.func = func;
	if(arg)
		n->write.arg = estrdup9p(arg);
	assert((n->dir.mode & DMDIR) == 0);
	n->dir.mode |= 0222;
}

void
createhook(Node *n, Putfunc func, char *arg, int isdir)
{

	n->write.func = func;
	n->create.isdir = isdir;
	if(arg)
		n->write.arg = estrdup9p(arg);
	assert((n->dir.mode & DMDIR) != 0);
	n->dir.mode |= 0222;
}

static Dir
protodir(Dir *parent, char *name, int mode, vlong length, 
	char *uid, char *gid, char *muid, long mtime, long atime)
{
	Dir d;
	long now = time(nil);

	if(uid == nil){
		if(parent)
			uid = parent->uid;
		if(uid == nil)
			uid = "jira";
	}
	
	if(gid == nil){
		if(parent)
			gid = parent->gid;
		if(gid == nil)
			gid = "jira";
	}

	if(muid == nil){
		if(parent)
			muid = parent->muid;
		if(muid == nil)
			muid = uid;
	}

	if(mtime == 0){
		if(parent)
			mtime = parent->mtime;
		if(mtime == 0)
			mtime = now;
	}

	if(atime == 0){
		if(parent)
			atime = parent->atime;
		if(atime == 0)
			atime = now;
	}

	/* cannot inherit or default these */
	d.name = estrdup9p(name);
	d.length = length;
	d.mode = mode;

	d.qid.type = (mode & DMDIR)? QTDIR: QTFILE;
	d.qid.path = hash(name);
	d.qid.vers = now;
	d.uid = estrdup9p(uid);
	d.gid = estrdup9p(gid);
	d.muid = estrdup9p(muid);
	d.mtime = mtime;
	d.atime = atime;

	return d;
}

Node *
newdir(Node **np, char *uid, char *gid, char *name)
{
	Node *t;
	Dir *proto;

	for(t = *np; t; t = t->next)
		if(strcmp(t->dir.name, name) == 0)
			break;
	if(t)
		return t;

	proto = nil;
	if(np && *np)
		proto = &(*np)->dir;

	t = emalloc9p(sizeof(Node));
	t->dir = protodir(proto, name, 0555|DMDIR, 0, uid, gid, nil, 0, 0);
	t->next = *np;
	*np = t;
	return t;
}

Node *
subdir(Node **np, char *name)
{
	Node *t;
	Dir *proto;

	for(t = *np; t; t = t->next)
		if(strcmp(t->dir.name, name) == 0)
			break;
	if(t)
		return t;

	proto = nil;
	if(np && *np)
		proto = &(*np)->dir;

	t = emalloc9p(sizeof(Node));
	t->dir = protodir(proto, name, 0755|DMDIR, 0, nil, nil, nil, 0, 0);
	t->next = *np;
	*np = t;
	return t;
}



static Node *
fullfile(Node **np, char *name, vlong size, char *uid, char *gid, char *muid, long mtime, long atime, char *buf, char *url)
{
	Node *t;
	Dir *proto;

	for(t = *np; t; t = t->next)
		if(strcmp(t->dir.name, name) == 0)
			break;
	if(t)
		return t;

	proto = nil;
	if(np && *np)
		proto = &(*np)->dir;

	t = emalloc9p(sizeof(Node));
	t->dir = protodir(proto, name, 0444, size, uid, gid, muid, mtime, atime);
	t->read.buf = buf;
	t->read.url = url;
	t->next = *np;
	*np = t;
	return t;
}

Node *
fulltext(Node **np, char *name, char *uid, char *gid, char *muid, long mtime, long atime, char *text)
{
	char *buf, *p;

	buf = emalloc9p(strlen(text)+2);
	strcpy(buf, text);

	/* exorcise ghosts of operating systems past */
	while((p = strchr(buf, '\r')) != nil)
		if(p[1] == '\n')
			*p = ' ';
		else
			*p = '\n';

	if((p = strrchr(text, '\n')) == nil || p[1] != 0)
		strcat(buf, "\n");

	return fullfile(np, name, strlen(buf), uid, gid, muid, mtime, atime, buf, nil);
}

Node *
fetchtext(Node **np, char *name, char *uid, char *gid, char *muid, long mtime, vlong size, char *url)
{
	return fullfile(np, name, size, uid, gid, muid, mtime, mtime, nil, estrdup9p(url));
}

Node *
subtext(Node **np, char *name, char *text)
{
	char *buf, *p;

	buf = emalloc9p(strlen(text)+2);
	strcpy(buf, text);

	/* exorcise gurls of operating systems past */
	while((p = strchr(buf, '\r')) != nil)
		if(p[1] == '\n')
			*p = ' ';
		else
			*p = '\n';

	if((p = strrchr(text, '\n')) == nil)
		p = text;
	if(p[1] != 0)
		strcat(buf, "\n");

	return fullfile(np, name, strlen(buf), nil, nil, nil, 0, 0, buf, nil);
}



static void
responderrstr(Req *r)
{
	char e[ERRMAX];

	*e = 0;
	rerrstr(e, sizeof(e));
	respond(r, e);
}

static void
fsattach(Req *r)
{
	Aux *a;
	char *spec = r->ifcall.aname;

	if(spec && *spec){
		respond(r, "invalid attach specifier");
		return;
	}

	a = emalloc9p(sizeof(Aux));
	r->fid->aux = a;
	a->node = Params.root;
	
	r->ofcall.qid = a->node->dir.qid;
	r->fid->qid = r->ofcall.qid;
	respond(r, nil);
}

static char*
fsclone(Fid *fid, Fid *nfid)
{
	Aux *a = fid->aux;
	Aux *na;

	na = emalloc9p(sizeof(Aux));
	na->node = a->node;
	nfid->aux = na;
	return nil;
}

static int
is_users(Node *n)
{
	if(n->parent != nil)
		return 0;
	if(strcmp(n->dir.name, "users") != 0)
		return 0;
	return 1;
}


static int
dirgen(int slot, Dir *d, void *aux)
{
	Node *n;
	Aux *a = aux;

	for(n = a->node->child; n; n = n->next)
		if(--slot < 0)
			break;
	if(n == nil)
		return -1;
	dircp(d, &n->dir);
	return 0;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	Aux *a = fid->aux;
	long now;
	Node *n;

	if(strcmp(name, "..") == 0){
		a->node = a->node->parent;
		fid->qid = a->node->dir.qid;
		*qid = a->node->dir.qid;
		return nil;
	}

	for(n = a->node->child; n; n = n->next){
		if(strcmp(n->dir.name, name) == 0){
			fid->qid = n->dir.qid;
			*qid = n->dir.qid;
			a->node->parent = n;
			a->node = n;
			return nil;
		}
	}

	/*
	 * auto-create the directory if we have a walk function,
	 * failing that do a directory scan.
	 */
	if(a->node->walk.func){
		if((n = (*a->node->walk.func)(a->node, Params.session, name)) != nil){
			fid->qid = n->dir.qid;
			*qid = n->dir.qid;
			a->node->parent = n;
			a->node = n;
			return nil;
		}
	}
	else
	if(a->node->scan.func){
		now = time(nil);
		if(a->node->scan.func && (! a->node->child || now - a->node->scan.last >= Params.cachetime)){
			a->node->child = (*a->node->scan.func)(a->node->child, Params.session, a->node->scan.arg);
			if(a->node->child)
				a->node->child->scan.last = now;
			for(n = a->node->child; n; n = n->next)
				if(strcmp(n->dir.name, name) == 0){
					fid->qid = n->dir.qid;
					*qid = n->dir.qid;
					a->node->parent = n;
					a->node = n;
					return nil;
			}
		}
	}

	if(Debug)
		fprint(2, "walk @%s -> %s failed\n", a->node->dir.name, name);
	return "not found";
}

static void
fsstat(Req *r)
{
	Aux *a = r->fid->aux;

	dircp(&r->d, &a->node->dir);
	respond(r, nil);
}

static void
fscreate(Req *r)
{
	Aux *a = r->fid->aux;

	if((a->node->dir.mode & 0222) == 0){
		respond(r, "dir is read only");
		return;
	}
	if((r->ifcall.perm & DMDIR) == 0 && a->node->create.isdir){
		respond(r, "dir create only");
		return;
	}
	if((r->ifcall.perm & DMDIR) != 0 && !a->node->create.isdir){
		respond(r, "file create only");
		return;
	}

	free(a->buf);
	a->len = 0;
	a->siz = 0;
	a->buf = nil;

	respond(r, nil);
}

static void
fsopen(Req *r)
{
	Aux *a = r->fid->aux;
	Node *n;
	long now;

	if (r->fid->qid.type & QTDIR){
		n = a->node;
		now = time(nil);
		if(n->scan.func && (! n->child || now - n->scan.last >= Params.cachetime)){
			n->child = (*n->scan.func)(n->child, Params.session, n->scan.arg);
			n->scan.last = now;
		}
		respond(r, nil);
		return;
	}
	if((r->ifcall.mode & OWRITE) == OWRITE){
		if((a->node->dir.mode & 0222) == 0){
			respond(r, "file read only");
			return;
		}
		free(a->buf);
		a->len = 0;
		a->siz = 0;
		a->buf = nil;
	}
	if((r->ifcall.mode & OREAD) == OREAD){
		if((a->node->dir.mode & 0444) == 0){
			respond(r, "file write only");
			return;
		}
		n = a->node;
		if(n->read.url){
			free(a->buf);
			a->buf = fetch(Params.session, n->read.url, n->dir.length);
			if(a->buf == nil){
				respond(r, "cannot fetch attachment");
				return;
			}
		}
		
	}
	respond(r, nil);
}

static void
fsread(Req *r)
{
	Aux *a = r->fid->aux;
	char *buf;
	Node *n;
	vlong off;
	long len, got;

	if (r->fid->qid.type & QTDIR){
		dirread9p(r, dirgen, a);
		respond(r, nil);
		return;
	}

	n = a->node;
	off = r->ifcall.offset;
	len = r->ifcall.count;
	got = n->dir.length;

	if(off >= got){
		r->ofcall.count = 0;
		respond(r, nil);
		return;
	}

	if(off+len >= got)
		len = got - off;

	if(n->read.url)
		buf = a->buf;
	else
		buf = n->read.buf;

	memmove(r->ofcall.data, buf+off, len);
	r->ofcall.count = len;
	respond(r, nil);
}

static void
fswrite(Req *r)
{
	Aux *a = r->fid->aux;
	vlong off;
	long len;

	off = r->ifcall.offset;
	len = r->ifcall.count;

	if(off+len+1 >= a->siz){		/* +1 to allow for terminating zero */
		a->siz = off+len+128;
		a->buf = erealloc9p(a->buf, a->siz);
		a->node->dir.length = off+len+1;
	}
	memmove(a->buf+off, r->ifcall.data, len);
	if(a->len < off+len){
		a->len = off+len;
		a->buf[a->len] = 0;
	}
	r->ofcall.count = len;
	respond(r, nil);
}

static void
fsdestroyfid(Fid *fid)
{
	Node *n;
	Aux *a = fid->aux;
	
	if(a == nil)
		return;
	n = a->node;

	if(n->write.func && a->buf && a->len)
		if((*n->write.func)(Params.session, n->write.arg, a->buf) < 0)
			fprint(2, "%s: %s write failed %r\n", argv0, n->dir.name);
	free(a->buf);
	free(a);
}
	
Srv fs =
{
	.destroyfid =	fsdestroyfid,
	.attach=	fsattach,
	.open=		fsopen,
	.create=	fscreate,
	.read=		fsread,
	.write=		fswrite,
	.stat=		fsstat,
	.clone= 	fsclone,
	.walk1= 	fswalk1,
};

static int
upperfmt(Fmt *f)
{
	int n;
	char *s, *p;

	s = va_arg(f->args, char *);
	n = 0;
	for(p = s; *p; p++)
		n += fmtprint(f, "%c", toupper(*p));
	return n;
}


static void
usage(void)
{
	fprint(2, "usage: %s [-a] [-c secs] [-w dir] [-m dir] [-s name] [url]\n", argv0);
	fprint(2, "  -a       don't authenticate\n");
	fprint(2, "  -c secs  filesystem cache time in secs (default 60)\n");
	fprint(2, "  -w dir   dir where webfs is mounted (default /mnt/web)\n");
	fprint(2, "  -m dir   dir where jira will be mounted (default /n/jira)\n");
	fprint(2, "  -s dir   name posted into /srv (none posted by default)\n");
	fprint(2, "  url     urlname of jira server.\n");
	fprint(2, "           if no url given then the $jira environment variable is used\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int tout;
	Node *c, *cc;
	Sess sess, *s;
	char *p, *mtpt, *svs;

	quotefmtinstall();
	fmtinstall('F', fcallfmt);
	fmtinstall('D', dirfmt);
	fmtinstall('U', upperfmt);

	s = &sess;
	memset(s, 0, sizeof(Sess));

	svs = nil;
	mtpt = nil;
	tout = 60;
	s->webfs = "/mnt/web";
	s->user = estrdup9p("anonymous");
	ARGBEGIN{
	case 'a':
		s->annon = 1;
		break;
	case 'c':
		tout = atoi(EARGF(usage()));
		break;
	case 'D':
		p = EARGF(usage());
		if(strstr(p, "9p"))
			chatty9p++;
		if(strstr(p, "json"))
			Dumpjson++;
		if(strstr(p, "tree"))
			Dumptree++;
		if(strstr(p, "time"))
			Dumptime++;
		if(strstr(p, "cache"))
			Dumpcache++;
		if(strstr(p, "valfail"))
			Dumpvalfail++;
		Debug = p;
		break;
	case 'w':
		s->webfs = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	case 's':
		svs = EARGF(usage());
		break;
	case 'M':
		mainmem->flags |= POOL_NOREUSE|POOL_PARANOIA|POOL_ANTAGONISM;
		break;
	default:
		usage();
	}ARGEND;

	if(argc == 1)
		s->url = argv[0];
	if(s->url == nil)
		if((s->url = getenv("jira")) == nil)
			sysfatal("jira url not set\n");

	if(strncmp(s->url, "http://", 7) != 0 &&
	   strncmp(s->url, "https://", 8) != 0)
		sysfatal("%s - not a valid url\n", s->url);

	if(dialweb(s) < 0)
		sysfatal("cannot get session to webfs\n");

	s->api = "latest";					/* why would we want anything less? */

	Params.root = nil;
	Params.session = s;
	Params.cachetime = tout;

	c = newdir(&Params.root, nil, nil, "");

	cc = newdir(&c->child, nil, nil, "projects");
	walkhook(cc, projvalid, nil);
	scanhook(cc, projlist, nil);

	cc = newdir(&c->child, nil, nil, "users");
	walkhook(cc, uservalid, nil);
	s->users = &cc->child;

	cc = newdir(&c->child, nil, nil, "session");
	scanhook(cc, sessioninfo, nil);
	sessioninfo(cc, s, nil);			/* call now to get our jira username */

	cc = newdir(&c->child, nil, nil, "server");
	scanhook(cc, serverinfo, nil);
	serverinfo(cc, s, nil);				/* call now to get the server version */

	switch(s->major){
	case 4:
		s->issueinfo = issueinfo_v4;
		break;
	case 5:
	case 6:
		s->issueinfo = issueinfo_v5;
		break;
	default: sysfatal("%d.x unknown server major version\n", s->major);
	}

	if(Dumptree)
		dumptree(Params.root, 0);

	if(!mtpt && !svs)
		mtpt = "/n/jira";

	postmountsrv(&fs, svs, mtpt, MREPL|MCREATE);
	exits(0);
}
